import math
from collections.abc import Mapping, Sequence
from os.path import join as pjoin
from ufo_model_loader.common import JSONLook  # type: ignore
from ufo_model_loader.commands import load_model, export_model, JSONLook  # type: ignore
from copy import deepcopy

USE_DETAILED_DIFF_COMPARISON = True

SENTINEL = object()


def compare_models(modelA, modelB):
    obj1_dict = dict(modelA.to_serializable_model().to_dict())
    obj2_dict = dict(modelB.to_serializable_model().to_dict())
    # from pprint import pprint
    # pprint(obj1_dict['parameters'])
    # pprint(obj2_dict['parameters'])
    if USE_DETAILED_DIFF_COMPARISON:
        diff = dict_diff(obj1_dict, obj2_dict)
        assert diff is None, f"Difference: {diff}"
    else:
        assert modelA.__dict__ == modelB.__dict__


def compare_dict_objects(obj1, obj2):
    obj1_dict = dict(obj1.__dict__)
    obj2_dict = dict(obj2.__dict__)
    # from pprint import pprint
    # pprint(obj1_dict)
    # pprint(obj2_dict)
    if USE_DETAILED_DIFF_COMPARISON:
        diff = dict_diff(obj1_dict, obj2_dict)
        assert diff is None, f"Difference: {diff}"
    else:
        assert obj1_dict == obj2_dict


def test_ufo_model_loader(tmp_path):

    # First loading the full model
    loaded_full_sm, input_param_card_full = load_model(
        input_model_path='sm',
        restriction_name='full',
        simplify_model=True,
    )
    assert loaded_full_sm is not None
    assert input_param_card_full is not None

    exported_model_path = export_model(
        model=loaded_full_sm,
        input_param_card=input_param_card_full,
        output_model_path=pjoin(tmp_path, 'sm_output_model_test.json'),
        json_look=JSONLook.COMPACT,
        allow_overwrite=True,
    )
    assert exported_model_path is not None

    # Now try re-loading the exported model again
    reloaded_sm_full, reloaded_param_card_full = load_model(
        input_model_path=exported_model_path,
        restriction_name='full',
        simplify_model=True,
    )
    assert reloaded_sm_full is not None
    assert reloaded_param_card_full is not None
    compare_models(reloaded_sm_full, loaded_full_sm)
    compare_dict_objects(reloaded_param_card_full, input_param_card_full)

    # Now test restrictions
    loaded_sm_no_b_mass_non_simplified, input_param_card_no_b_mass_non_simplified = load_model(
        input_model_path='sm',
        restriction_name='no_b_mass',
        simplify_model=False,
    )
    assert loaded_sm_no_b_mass_non_simplified is not None
    assert input_param_card_no_b_mass_non_simplified is not None

    with open(pjoin(tmp_path, 'restrict_no_b_mass.json'), 'w', encoding='utf-8') as f:
        f.write(input_param_card_no_b_mass_non_simplified.to_json(JSONLook.VERBOSE))

    loaded_sm_no_b_mass, input_param_card_no_b_mass = load_model(
        input_model_path='sm',
        restriction_name='no_b_mass',
        simplify_model=True,
    )
    assert loaded_sm_no_b_mass is not None
    assert input_param_card_no_b_mass is not None

    re_loaded_sm_no_b_mass, re_loaded_input_param_card_no_b_mass = load_model(
        input_model_path=pjoin(tmp_path, 'sm_output_model_test.json'),
        restriction_name='no_b_mass',
        simplify_model=True,
    )
    assert re_loaded_sm_no_b_mass is not None
    assert re_loaded_input_param_card_no_b_mass is not None

    compare_models(re_loaded_sm_no_b_mass, loaded_sm_no_b_mass)
    compare_dict_objects(
        re_loaded_input_param_card_no_b_mass, input_param_card_no_b_mass)


def dict_diff(a, b, *, path="root", rel_tol=None, abs_tol=None):
    """Return None if a == b (deep), else a string describing the first difference.

    - dicts: missing/extra keys, then recurse into shared keys
    - lists/tuples: length mismatch, then recurse by index
    - sets/frozensets: report first element only-in-one
    - floats: optional tolerance via float_tol
    - treats NaNs as equal if both are NaN
    """

    # Fast path for exact equality including identical objects
    if a is b:
        return None

    # Handle numeric tolerance and NaNs
    if _both_numbers(a, b):
        if _nums_equal(a, b, rel_tol=rel_tol, abs_tol=abs_tol):
            return None
        return f"{path}: {a!r} != {b!r}"

    # Type mismatch (after numeric normalization)
    if type(a) is not type(b):
        # Allow tuple vs list? No: be strict to catch bugs.
        return f"{path}: type mismatch {type(a).__name__} != {type(b).__name__}"

    # Dict-like
    if isinstance(a, Mapping):
        # missing in b
        for k in sorted(set(a) - set(b), key=_kkey):
            return f"{path}[{k!r}] only in a"
        # missing in a
        for k in sorted(set(b) - set(a), key=_kkey):
            return f"{path}[{k!r}] only in b"
        # recurse shared keys in stable order
        for k in sorted(a.keys(), key=_kkey):
            d = dict_diff(a[k], b.get(k, SENTINEL), path=f"{
                          path}[{k!r}]", rel_tol=rel_tol, abs_tol=abs_tol)
            if d:
                return d
        return None

    # Sequence (but not str/bytes)
    if isinstance(a, Sequence) and not isinstance(a, (str, bytes, bytearray)):
        if len(a) != len(b):
            return f"{path}: length {len(a)} != {len(b)}"
        for i, (ai, bi) in enumerate(zip(a, b)):
            d = dict_diff(ai, bi, path=f"{
                          path}[{i}]", rel_tol=rel_tol, abs_tol=abs_tol)
            if d:
                return d
        return None

    # Sets
    if isinstance(a, (set, frozenset)):
        if a == b:
            return None
        only_a = sorted(a - b, key=_ekey)
        if only_a:
            return f"{path}: element only in a -> {only_a[0]!r}"
        only_b = sorted(b - a, key=_ekey)
        if only_b:
            return f"{path}: element only in b -> {only_b[0]!r}"
        # Fallback
        return f"{path}: set contents differ"

    # Bytes/bytearray exact compare
    if isinstance(a, (bytes, bytearray)):
        if a == b:
            return None
        return f"{path}: bytes differ (len {len(a)} != {len(b)})" if len(a) != len(b) else f"{path}: bytes differ"

    # Fallback: plain equality
    if a == b:
        return None
    return f"{path}: {a!r} != {b!r}"


def _both_numbers(a, b):
    return isinstance(a, (int, float)) and isinstance(b, (int, float))


def _nums_equal(a, b, *, rel_tol, abs_tol):
    if isinstance(a, float) and isinstance(b, float) and math.isnan(a) and math.isnan(b):
        return True
    if rel_tol is not None or abs_tol is not None:
        return math.isclose(float(a), float(b),
                            rel_tol=0.0 if rel_tol is None else rel_tol,
                            abs_tol=0.0 if abs_tol is None else abs_tol)
    return a == b


def _kkey(k):
    # Sort keys deterministically even if mixed types
    return (str(type(k)), repr(k))


def _ekey(e):
    return (str(type(e)), repr(e))


def test_equal_simple_dicts():
    A = {"a": 1, "b": 2}
    B = {"a": 1, "b": 2}
    assert dict_diff(A, B) is None


def test_missing_key_in_a():
    A = {"a": 1}
    B = {"a": 1, "b": 2}
    msg = dict_diff(A, B)
    assert "only in b" in msg
    assert "root['b']" in msg


def test_missing_key_in_b():
    A = {"a": 1, "b": 2}
    B = {"a": 1}
    msg = dict_diff(A, B)
    assert "only in a" in msg
    assert "root['b']" in msg


def test_nested_difference():
    A = {"a": {"x": 1}}
    B = {"a": {"x": 2}}
    msg = dict_diff(A, B)
    assert msg.startswith("root['a']['x']")


def test_sequence_length_mismatch():
    A = {"a": [1, 2]}
    B = {"a": [1, 2, 3]}
    msg = dict_diff(A, B)
    assert "length" in msg
    assert "root['a']" in msg


def test_sequence_value_mismatch():
    A = [1, 2, 3]
    B = [1, 2, 4]
    msg = dict_diff(A, B)
    assert msg.startswith("root[2]")


def test_set_difference():
    A = {"s": {1, 2}}
    B = {"s": {1, 3}}
    msg = dict_diff(A, B)
    assert "element only" in msg
    assert "root['s']" in msg


def test_type_mismatch():
    A = {"x": [1, 2]}
    B = {"x": (1, 2)}
    msg = dict_diff(A, B)
    assert "type mismatch" in msg
    assert "list" in msg and "tuple" in msg


def test_float_exact_equal():
    A = {"x": 1.0}
    B = {"x": 1.0}
    assert dict_diff(A, B) is None


def test_float_within_abs_tol():
    A = {"x": 1.0001}
    B = {"x": 1.0002}
    assert dict_diff(A, B, abs_tol=1e-3) is None


def test_float_outside_abs_tol():
    A = {"x": 1.0}
    B = {"x": 1.1}
    msg = dict_diff(A, B, abs_tol=1e-3)
    assert msg.startswith("root['x']")


def test_float_nan_equal():
    A = {"x": float("nan")}
    B = {"x": float("nan")}
    assert dict_diff(A, B) is None


def test_bytes_difference():
    A = {"x": b"abc"}
    B = {"x": b"abd"}
    msg = dict_diff(A, B)
    assert "bytes differ" in msg


def test_complex_nested_structure():
    A = {"a": [1, {"b": (2, 3)}]}
    B = {"a": [1, {"b": (2, 4)}]}
    msg = dict_diff(A, B)
    assert msg == "root['a'][1]['b'][1]: 3 != 4"



# type: ignore
# This file was automatically created by FeynRules 1.7.69
# Mathematica version: 8.0 for Mac OS X x86 (64-bit) (November 6, 2010)
# Date: Mon 1 Oct 2012 14:58:26


from __future__ import absolute_import
from .object_library import all_orders, CouplingOrder


QCD = CouplingOrder(name='QCD',
                    expansion_order=99,
                    hierarchy=1)

QED = CouplingOrder(name='QED',
                    expansion_order=99,
                    hierarchy=2)



# type: ignore
##
##
# Feynrules Header
##
##
##
##
##

from __future__ import absolute_import
import cmath


class UFOBaseClass(object):
    """The class from which all FeynRules classes are derived."""

    require_args = []

    def __init__(self, *args, **options):
        assert (len(self.require_args) == len(args))

        for i, name in enumerate(self.require_args):
            setattr(self, name, args[i])

        for (option, value) in options.items():
            setattr(self, option, value)

    def get(self, name):
        return getattr(self, name)

    def set(self, name, value):
        setattr(self, name, value)

    def get_all(self):
        """Return a dictionary containing all the information of the object"""
        return self.__dict__

    def __str__(self):
        return self.name

    def nice_string(self):
        """ return string with the full information """
        return '\n'.join(['%s \t: %s' % (name, value) for name, value in self.__dict__.items()])

    def __repr__(self):
        replacements = [
            ('+', '__plus__'),
            ('-', '__minus__'),
            ('@', '__at__'),
            ('!', '__exclam__'),
            ('?', '__quest__'),
            ('*', '__star__'),
            ('~', '__tilde__')
        ]
        text = self.name
        for orig, sub in replacements:
            text = text.replace(orig, sub)
        return text


all_particles = []


class Particle(UFOBaseClass):
    """A standard Particle"""

    require_args = ['pdg_code', 'name', 'antiname', 'spin',
                    'color', 'mass', 'width', 'texname', 'antitexname', 'charge']

    require_args_all = ['pdg_code', 'name', 'antiname', 'spin', 'color', 'mass', 'width',
                        'texname', 'antitexname', 'charge', 'line', 'propagating', 'goldstoneboson']

    def __init__(self, pdg_code, name, antiname, spin, color, mass, width, texname,
                 antitexname, charge, line=None, propagating=True, goldstoneboson=False, **options):

        args = (pdg_code, name, antiname, spin, color, mass, width, texname,
                antitexname, float(charge))

        UFOBaseClass.__init__(self, *args,  **options)

        global all_particles
        all_particles.append(self)

        self.propagating = propagating
        self.goldstoneboson = goldstoneboson

        self.selfconjugate = (name == antiname)
        if 1:  # not line:
            self.line = self.find_line_type()
        else:
            self.line = line

    def find_line_type(self):
        """ find how we draw a line if not defined
        valid output: dashed/straight/wavy/curly/double/swavy/scurly
        """

        spin = self.spin
        color = self.color

        # use default
        if spin == 1:
            return 'dashed'
        elif spin == 2:
            if not self.selfconjugate:
                return 'straight'
            elif color == 1:
                return 'swavy'
            else:
                return 'scurly'
        elif spin == 3:
            if color == 1:
                return 'wavy'

            else:
                return 'curly'
        elif spin == 5:
            return 'double'
        elif spin == -1:
            return 'dotted'
        else:
            return 'dashed'  # not supported yet

    def anti(self):
        if self.selfconjugate:
            raise Exception('%s has no anti particle.' % self.name)
        outdic = {}
        for k, v in self.__dict__.items():
            if k not in self.require_args_all:
                outdic[k] = -v
        if self.color in [1, 8]:
            newcolor = self.color
        else:
            newcolor = -self.color

        return Particle(-self.pdg_code, self.antiname, self.name, self.spin, newcolor, self.mass, self.width,
                        self.antitexname, self.texname, -self.charge, self.line, self.propagating, self.goldstoneboson, **outdic)


all_parameters = []


class Parameter(UFOBaseClass):

    require_args = ['name', 'nature', 'type', 'value', 'texname']

    def __init__(self, name, nature, type, value, texname, lhablock=None, lhacode=None):

        args = (name, nature, type, value, texname)

        UFOBaseClass.__init__(self, *args)

        args = (name, nature, type, value, texname)

        global all_parameters
        all_parameters.append(self)

        if (lhablock is None or lhacode is None) and nature == 'external':
            raise Exception(
                'Need LHA information for external parameter "%s".' % name)
        self.lhablock = lhablock
        self.lhacode = lhacode


all_vertices = []


class Vertex(UFOBaseClass):

    require_args = ['name', 'particles', 'color', 'lorentz', 'couplings']

    def __init__(self, name, particles, color, lorentz, couplings, **opt):

        args = (name, particles, color, lorentz, couplings)

        UFOBaseClass.__init__(self, *args, **opt)

        args = (particles, color, lorentz, couplings)

        global all_vertices
        all_vertices.append(self)


all_couplings = []


class Coupling(UFOBaseClass):

    require_args = ['name', 'value', 'order']

    def __init__(self, name, value, order, **opt):

        args = (name, value, order)
        UFOBaseClass.__init__(self, *args, **opt)
        global all_couplings
        all_couplings.append(self)


all_lorentz = []


class Lorentz(UFOBaseClass):

    require_args = ['name', 'spins', 'structure']

    def __init__(self, name, spins, structure='external', **opt):
        args = (name, spins, structure)
        UFOBaseClass.__init__(self, *args, **opt)

        global all_lorentz
        all_lorentz.append(self)


all_functions = []


class Function(object):

    def __init__(self, name, arguments, expression):

        global all_functions
        all_functions.append(self)

        self.name = name
        self.arguments = arguments
        self.expr = expression

    def __call__(self, *opt):

        for i, arg in enumerate(self.arguments):
            exec('%s = %s' % (arg, opt[i]))

        return eval(self.expr)


all_orders = []


class CouplingOrder(object):

    def __init__(self, name, expansion_order, hierarchy, perturbative_expansion=0):

        global all_orders
        all_orders.append(self)

        self.name = name
        self.expansion_order = expansion_order
        self.hierarchy = hierarchy


all_decays = []


class Decay(UFOBaseClass):
    require_args = ['particle', 'partial_widths']

    def __init__(self, particle, partial_widths, **opt):
        args = (particle, partial_widths)
        UFOBaseClass.__init__(self, *args, **opt)

        global all_decays
        all_decays.append(self)

        # Add the information directly to the particle
        particle.partial_widths = partial_widths


all_form_factors = []


class FormFactor(UFOBaseClass):
    require_args = ['name', 'type', 'value']

    def __init__(self, name, type, value, **opt):
        args = (name, type, value)
        UFOBaseClass.__init__(self, *args, **opt)

        global all_form_factors
        all_form_factors.append(self)



# type: ignore
from __future__ import absolute_import
from . import particles
from . import couplings
from . import lorentz
from . import parameters
from . import vertices
from . import coupling_orders
from . import function_library
from . import object_library

gauge = [0]

all_particles = particles.all_particles
all_vertices = vertices.all_vertices
all_couplings = couplings.all_couplings
all_lorentz = lorentz.all_lorentz
all_parameters = parameters.all_parameters
all_orders = coupling_orders.all_orders
all_functions = function_library.all_functions


__author__ = "Valentin Hirschi"
__version__ = "1.0"
__email__ = "valentin.hirschi@gmail.com"



import os
import logging
import sys
import re

from importlib.resources import files, as_file
from pathlib import Path
from typing import Any, TextIO
from enum import StrEnum, Enum
from pprint import pformat
import json


def optionally_lower_external_parameter_name(name: str) -> str:
    # return name.lower()
    return name


JSBEAUTIFIER_AVAILABLE = True
try:
    import jsbeautifier  # type: ignore
except ModuleNotFoundError:
    JSBEAUTIFIER_AVAILABLE = False

ROOT_PATH = Path()
try:
    with as_file(files("ufo_model_loader")) as p:
        ROOT_PATH = p
except ModuleNotFoundError:
    ROOT_PATH = Path(os.path.abspath(os.path.join(
        os.path.dirname(os.path.realpath(__file__)), os.path.pardir, 'ufo_model_loader')))

DATA_PATH = os.path.abspath(os.path.join(ROOT_PATH, 'data'))

logger = logging.getLogger('UFOModelLoader')


class JSONLook(StrEnum):
    VERBOSE = 'verbose'
    PRETTY = 'pretty'
    COMPACT = 'compact'


class UFOModelLoaderError(Exception):
    pass


class UFOModelLoaderWarning(Enum):
    FloatInExpression = 100
    DroppingEpsilonTerms = 101

    def __str__(self):
        if self == UFOModelLoaderWarning.FloatInExpression:
            return "FloatInExpression"
        elif self == UFOModelLoaderWarning.DroppingEpsilonTerms:
            return "DroppingEpsilonTerms"
        else:
            raise UFOModelLoaderError(f"Unknown side: {self}")


UFOMODELLOADER_WARNINGS_ISSUED: set[UFOModelLoaderWarning] = set()


class Colour(StrEnum):
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    GRAY = '\033[21m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'


class UFOModelLoaderCustomFormatter(logging.Formatter):
    """Logging colored formatter"""

    def __init__(self, fmt: str, datefmt: str | None = None):
        super().__init__(datefmt=datefmt)
        self.fmt = fmt
        self.datefmt = datefmt

    def format(self, record: logging.LogRecord) -> str:
        formatter = logging.Formatter(self.fmt, self.datefmt)
        if record.levelno != logging.DEBUG:
            if record.name.startswith('_gammaloop'):
                record.name = f"rust.{record.name[11:]}"
            if len(record.name) > 20:
                record.name = f"{record.name[:17]}..."
            record.name = f"{record.name:20}"
        match record.levelno:
            case logging.DEBUG:
                record.levelname = f"{Colour.GRAY}{record.levelname:8}{Colour.END}"  # nopep8
            case logging.INFO:
                record.levelname = f"{record.levelname:8}"  # nopep8
            case logging.WARNING:
                record.levelname = f"{Colour.YELLOW}{record.levelname:8}{Colour.END}"  # nopep8
            case logging.ERROR:
                record.levelname = f"{Colour.RED}{record.levelname:8}{Colour.END}"  # nopep8
            case logging.CRITICAL:
                record.levelname = f"{Colour.RED}{Colour.BOLD}{record.levelname:8}{Colour.END}"  # nopep8
            case _:
                record.levelname = f"{record.levelname:8}"
        record.asctime = self.formatTime(record, self.datefmt)
        return formatter.format(record)


def setup_logging(prefix_format) -> logging.StreamHandler[TextIO]:
    match prefix_format:
        case 'none':
            console_format = f'%(message)s'
            time_format = "%H:%M:%S"
        case 'min':
            console_format = f'%(levelname)s: %(message)s'
            time_format = "%H:%M:%S"
        case 'short':
            console_format = f'[{Colour.GREEN}%(asctime)s{Colour.END}] %(levelname)s: %(message)s'  # nopep8
            time_format = "%H:%M:%S"
        case 'long':
            console_format = f'[{Colour.GREEN}%(asctime)s.%(msecs)03d{Colour.END}] @{Colour.BLUE}%(name)s{Colour.END} %(levelname)s: %(message)s'  # nopep8
            time_format = '%Y-%m-%d %H:%M:%S'
        case _:
            raise UFOModelLoaderError(
                "Invalid prefix_format: %s", prefix_format)
    file_format = '[%(asctime)s] %(name)s %(levelname)s: %(message)s'
    console_formatter = UFOModelLoaderCustomFormatter(
        console_format, datefmt=time_format)
    file_formatter = UFOModelLoaderCustomFormatter(file_format)
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.INFO)

    console_handler.setFormatter(console_formatter)
    logging.getLogger().handlers = []
    logging.getLogger().addHandler(console_handler)

    if 'UFO_MODEL_LOADER_ENABLE_FILE_HANDLERS' in os.environ and os.environ['UFO_MODEL_LOADER_ENABLE_FILE_HANDLERS'].upper() != 'FALSE':
        log_file_name = 'ufo_model_loader_debug.log'
        file_handler = logging.FileHandler(log_file_name, mode='w')
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(file_formatter)
        logging.getLogger().addHandler(file_handler)

        error_file_name = 'ufo_model_loader_error.log'
        error_file_handler = logging.FileHandler(error_file_name, mode='w')
        error_file_handler.setLevel(logging.ERROR)
        error_file_handler.setFormatter(file_formatter)
        logging.getLogger().addHandler(error_file_handler)

    logging.getLogger().setLevel(logging.DEBUG)

    return console_handler


def verbose_json_dump(obj: Any, json_look: JSONLook = JSONLook.VERBOSE) -> str:
    match json_look:
        case JSONLook.COMPACT:
            return json.dumps(obj, sort_keys=True, ensure_ascii=False)
        case JSONLook.VERBOSE:
            return json.dumps(obj, indent=2, sort_keys=True, ensure_ascii=False)
        case JSONLook.PRETTY:
            if not JSBEAUTIFIER_AVAILABLE:
                raise UFOModelLoaderError(
                    "JSBeautifier not available, cannot use 'pretty' JSON look. Install this dependencies with e.g. `pip install jsbeautifier`.")
            json_dump = json.dumps(obj, sort_keys=True, ensure_ascii=False)
            jsbeautifier_options = jsbeautifier.default_options()
            jsbeautifier_options.indent_size = 2
            return jsbeautifier.beautify(json_dump, jsbeautifier_options)
        case _:
            raise UFOModelLoaderError(
                f"JSON look selected not supported: {json_look}")



# type: ignore
from __future__ import absolute_import
from .object_library import all_couplings, Coupling


SCALAR_COUPLING = Coupling(name='SCALAR_COUPLING',
                           value='lam*complex(0,1)',
                           order={'QCD': 1})



# type: ignore
from __future__ import absolute_import
from . import particles
from . import couplings
from . import lorentz
from . import parameters
from . import vertices
from . import coupling_orders
from . import function_library
from . import object_library
# from . import write_param_card
try:
    from . import decays
except ImportError:
    pass
try:
    from . import build_restrict
except ImportError:
    pass

# model options
gauge = [0, 1]


all_particles = particles.all_particles
all_vertices = vertices.all_vertices
all_couplings = couplings.all_couplings
all_lorentz = lorentz.all_lorentz
all_parameters = parameters.all_parameters
all_orders = coupling_orders.all_orders
all_functions = function_library.all_functions
# all_decays = decays.all_decays


__author__ = "N. Christensen, C. Duhr"
__version__ = "1.3"
__email__ = "neil@pa.msu.edu, claude.duhr@uclouvain.be"



import os
from ufo_model_loader.model import Model, InputParamCard, ParamCard  # type: ignore
from ufo_model_loader.common import DATA_PATH, UFOModelLoaderError, UFOMODELLOADER_WARNINGS_ISSUED, JSONLook, logger, Colour  # type: ignore
from os.path import join as pjoin


def load_model(input_model_path: str, restriction_name: str | None, simplify_model: bool, wrap_indices_in_lorentz_structures: bool) -> tuple[Model, InputParamCard]:

    INPUT_FORMAT = 'JSON' if input_model_path.upper().endswith('.JSON') else 'UFO'

    # Derive the full model path from possible known locations
    if os.path.isabs(input_model_path):
        if not os.path.exists(input_model_path):
            raise UFOModelLoaderError(
                f"Absolute {INPUT_FORMAT} model path {input_model_path} not found. Please check the path.")
        model_base_path = os.path.dirname(input_model_path)
        model_name = os.path.basename(input_model_path)
    else:
        model_full_path = os.path.abspath(pjoin(os.getcwd(), input_model_path))
        if not os.path.exists(model_full_path):
            model_full_path = os.path.abspath(
                pjoin(DATA_PATH, 'models', input_model_path))
            if not os.path.exists(model_full_path):
                raise UFOModelLoaderError(
                    f"Relative {INPUT_FORMAT} model path/name {input_model_path} not found in current working directory or in '{pjoin(DATA_PATH, 'models')}'. Considering providing an absolute path.")
        model_base_path = os.path.dirname(model_full_path)
        model_name = os.path.basename(model_full_path)

    logger.info("Loading %s%s%s model %s%s%s from directory '%s%s%s'",
                Colour.BLUE, INPUT_FORMAT, Colour.END, Colour.GREEN, model_name, Colour.END, Colour.BLUE, model_base_path, Colour.END)

    # Try and load the model
    if INPUT_FORMAT == 'JSON':
        with open(pjoin(model_base_path, model_name), 'r', encoding='utf-8') as f:
            model = Model.from_json(f.read())
    else:
        model = Model.from_ufo_model(pjoin(model_base_path, model_name))

    # Make sure to issue warnings again if they were issued before
    UFOMODELLOADER_WARNINGS_ISSUED.clear()

    # Identify the restriction file path
    restriction_card: InputParamCard | None = None
    if restriction_name is not None:
        if restriction_name.lower() not in ['full', 'none']:
            if INPUT_FORMAT == 'UFO':
                if not os.path.isfile(pjoin(model_base_path, model_name, f'restrict_{restriction_name}.dat')):
                    raise UFOModelLoaderError(
                        f"Restriction file 'restrict_{restriction_name}.dat' not found for UFO model '{model_name}' in directory '{pjoin(model_base_path, model_name)}'.")
                param_card = ParamCard(
                    pjoin(model_base_path, model_name, f'restrict_{restriction_name}.dat'))
                restriction_card = InputParamCard.from_param_card(
                    param_card=param_card, model=model)
            else:
                if not os.path.isfile(pjoin(model_base_path, f'restrict_{restriction_name}.json')):
                    raise UFOModelLoaderError(
                        f"Restriction file 'restrict_{restriction_name}.json' not found for JSON model '{model_name}' in directory '{model_base_path}'.")
                restriction_card = InputParamCard.from_json_file(
                    json_path=pjoin(model_base_path, f'restrict_{restriction_name}.json'))
    else:
        if INPUT_FORMAT == 'UFO':
            if os.path.isfile(pjoin(model_base_path, model_name, 'restrict_default.dat')):
                param_card = ParamCard(
                    pjoin(model_base_path, model_name, 'restrict_default.dat'))
                restriction_card = InputParamCard.from_param_card(
                    param_card=param_card, model=model)
        else:
            if os.path.isfile(pjoin(model_base_path, f'{model_name}_default.json')):
                restriction_card = InputParamCard.from_json_file(
                    json_path=pjoin(model_base_path, f'{model_name}_default.json'))

    if restriction_card is not None:
        if restriction_name is not None:
            logger.info(
                "Applying restriction %s%s%s to model %s%s%s", Colour.BLUE, restriction_name, Colour.END, Colour.GREEN, model_name, Colour.END)
        else:
            logger.info(
                "Applying %sdefault%s restriction to model %s%s%s", Colour.BLUE, Colour.END, Colour.GREEN, model_name, Colour.END)
        if not simplify_model:
            logger.warning(
                "Model simplification from restriction %sdisabled%s, so no removal of zero contributions will be performed.", Colour.RED, Colour.END)
        model.apply_input_param_card(restriction_card, simplify=simplify_model)
    else:
        logger.info("No restriction applied to model '%s%s%s'",
                    Colour.GREEN, model_name, Colour.END)

    if wrap_indices_in_lorentz_structures:
        model.wrap_indices_in_lorentz_structures()
        logger.info("Wrapped indices in Lorentz structures for model %s%s%s",
                    Colour.GREEN, model_name, Colour.END)

    input_param_card = InputParamCard.from_model(model)
    if restriction_card is not None:
        for param in input_param_card.keys():
            if param in restriction_card:
                input_param_card[param] = restriction_card[param]

    logger.info("Model %s%s%s successfully loaded (%s%s%s particles, %s%s%s parameters, %s%s%s interactions, %s%s%s couplings, %s%s%s Lorentz structures)",
                Colour.GREEN, model.name, Colour.END,
                Colour.YELLOW, len(model.particles), Colour.END,
                Colour.YELLOW, len(model.parameters), Colour.END,
                Colour.YELLOW, len(model.vertex_rules), Colour.END,
                Colour.YELLOW, len(model.couplings), Colour.END,
                Colour.YELLOW, len(model.lorentz_structures), Colour.END
                )
    return model, input_param_card


def export_model(model: Model, input_param_card: InputParamCard, output_model_path: str | None = None, json_look: JSONLook = JSONLook.VERBOSE, allow_overwrite=False) -> str | None:

    if output_model_path is None:
        output_model_path = os.path.abspath(
            pjoin(os.getcwd(), f"{model.name}.json"))

    if not output_model_path.upper().endswith('.JSON'):
        raise UFOModelLoaderError(
            f"Output model path '{output_model_path}' must end with .json as this is the only format supported for output.")
    model_base_path = os.path.dirname(os.path.abspath(output_model_path))
    model_output_name = os.path.basename(output_model_path[:-5])

    model_output_path = pjoin(model_base_path, f"{model_output_name}.json")
    input_param_card_output_path = pjoin(
        model_base_path, f"{model_output_name}_param_card.json")
    if os.path.isfile(model_output_path) and not allow_overwrite:
        logger.exception(
            "Output model file '%s%s%s' already exists. Use option %s-w%s to allow overwriting it.", Colour.RED, input_param_card_output_path, Colour.END, Colour.GREEN, Colour.END)
        return None
    if os.path.isfile(input_param_card_output_path) and not allow_overwrite:
        logger.critical(
            "Output input parameter card file '%s%s%s' already exists. Use option %s-w%s to allow overwriting it.",
            Colour.RED, input_param_card_output_path, Colour.END, Colour.GREEN, Colour.END)
        return None

    with open(model_output_path, 'w', encoding='utf-8') as f:
        f.write(model.to_json(json_look=json_look))
    with open(input_param_card_output_path, 'w', encoding='utf-8') as f:
        f.write(input_param_card.to_json(json_look=json_look))
    if os.getcwd() == model_base_path:
        logger.info(
            "Successfully exported model in %s%s%s %sJSON%s format to file '%s%s.json%s' and corresponding input parameter card to '%s%s_param_card.json%s'", Colour.YELLOW, json_look, Colour.END, Colour.BLUE, Colour.END, Colour.GREEN, model_output_name, Colour.END, Colour.GREEN, model_output_name, Colour.END)
    else:
        logger.info("Successfully exported model in %s%s%s %sJSON%s format to files '%s%s.json%s' and corresponding input parameter card to '%s%s%s' in directory '%s%s%s'",
                    Colour.YELLOW, json_look, Colour.END, Colour.BLUE, Colour.END, Colour.GREEN, f"{model_output_name}.json", Colour.END, Colour.GREEN, f"{model_output_name}_param_card.json", Colour.END, Colour.GREEN, model_base_path, Colour.END)

    return model_output_path



# type: ignore
from __future__ import absolute_import
from .object_library import all_lorentz, Lorentz

from .function_library import complexconjugate, re, im, csc, sec, acsc, asec
from . import parameters as Param

try:
    import form_factors as ForFac
except ImportError:
    pass

scalar_lorentz_structures = {}
for n_point_interaction in Param.N_POINT_INTERACTIONS:
    scalar_lorentz_structures[n_point_interaction] = Lorentz(name='SCALAR_%d_LORENTZ_STRUCTURE' % n_point_interaction,
                                                             spins=[
                                                                 1, ]*n_point_interaction,
                                                             structure='1')



# This file was automatically created by FeynRules 1.7.69
# Mathematica version: 8.0 for Mac OS X x86 (64-bit) (November 6, 2010)
# Date: Mon 1 Oct 2012 14:58:25


from __future__ import absolute_import
from .object_library import all_vertices, Vertex
from . import particles as P
from . import couplings as C
from . import lorentz as L


V_1 = Vertex(name = 'V_1',
             particles = [ P.G0, P.G0, P.G0, P.G0 ],
             color = [ '1' ],
             lorentz = [ L.SSSS1 ],
             couplings = {(0,0):C.GC_33})

V_2 = Vertex(name = 'V_2',
             particles = [ P.G0, P.G0, P.G__minus__, P.G__plus__ ],
             color = [ '1' ],
             lorentz = [ L.SSSS1 ],
             couplings = {(0,0):C.GC_31})

V_3 = Vertex(name = 'V_3',
             particles = [ P.G__minus__, P.G__minus__, P.G__plus__, P.G__plus__ ],
             color = [ '1' ],
             lorentz = [ L.SSSS1 ],
             couplings = {(0,0):C.GC_32})

V_4 = Vertex(name = 'V_4',
             particles = [ P.G0, P.G0, P.H, P.H ],
             color = [ '1' ],
             lorentz = [ L.SSSS1 ],
             couplings = {(0,0):C.GC_31})

V_5 = Vertex(name = 'V_5',
             particles = [ P.G__minus__, P.G__plus__, P.H, P.H ],
             color = [ '1' ],
             lorentz = [ L.SSSS1 ],
             couplings = {(0,0):C.GC_31})

V_6 = Vertex(name = 'V_6',
             particles = [ P.H, P.H, P.H, P.H ],
             color = [ '1' ],
             lorentz = [ L.SSSS1 ],
             couplings = {(0,0):C.GC_33})

V_7 = Vertex(name = 'V_7',
             particles = [ P.G0, P.G0, P.H ],
             color = [ '1' ],
             lorentz = [ L.SSS1 ],
             couplings = {(0,0):C.GC_68})

V_8 = Vertex(name = 'V_8',
             particles = [ P.G__minus__, P.G__plus__, P.H ],
             color = [ '1' ],
             lorentz = [ L.SSS1 ],
             couplings = {(0,0):C.GC_68})

V_9 = Vertex(name = 'V_9',
             particles = [ P.H, P.H, P.H ],
             color = [ '1' ],
             lorentz = [ L.SSS1 ],
             couplings = {(0,0):C.GC_69})

V_10 = Vertex(name = 'V_10',
              particles = [ P.a, P.a, P.G__minus__, P.G__plus__ ],
              color = [ '1' ],
              lorentz = [ L.VVSS1 ],
              couplings = {(0,0):C.GC_6})

V_11 = Vertex(name = 'V_11',
              particles = [ P.a, P.G__minus__, P.G__plus__ ],
              color = [ '1' ],
              lorentz = [ L.VSS1 ],
              couplings = {(0,0):C.GC_3})

V_12 = Vertex(name = 'V_12',
              particles = [ P.ghA, P.ghWm__tilde__, P.W__minus__ ],
              color = [ '1' ],
              lorentz = [ L.UUV1 ],
              couplings = {(0,0):C.GC_3})

V_13 = Vertex(name = 'V_13',
              particles = [ P.ghA, P.ghWp__tilde__, P.W__plus__ ],
              color = [ '1' ],
              lorentz = [ L.UUV1 ],
              couplings = {(0,0):C.GC_4})

V_14 = Vertex(name = 'V_14',
              particles = [ P.ghWm, P.ghA__tilde__, P.G__plus__ ],
              color = [ '1' ],
              lorentz = [ L.UUS1 ],
              couplings = {(0,0):C.GC_75})

V_15 = Vertex(name = 'V_15',
              particles = [ P.ghWm, P.ghA__tilde__, P.W__plus__ ],
              color = [ '1' ],
              lorentz = [ L.UUV1 ],
              couplings = {(0,0):C.GC_3})

V_16 = Vertex(name = 'V_16',
              particles = [ P.ghWm, P.ghWm__tilde__, P.G0 ],
              color = [ '1' ],
              lorentz = [ L.UUS1 ],
              couplings = {(0,0):C.GC_70})

V_17 = Vertex(name = 'V_17',
              particles = [ P.ghWm, P.ghWm__tilde__, P.H ],
              color = [ '1' ],
              lorentz = [ L.UUS1 ],
              couplings = {(0,0):C.GC_71})

V_18 = Vertex(name = 'V_18',
              particles = [ P.ghWm, P.ghWm__tilde__, P.a ],
              color = [ '1' ],
              lorentz = [ L.UUV1 ],
              couplings = {(0,0):C.GC_4})

V_19 = Vertex(name = 'V_19',
              particles = [ P.ghWm, P.ghWm__tilde__, P.Z ],
              color = [ '1' ],
              lorentz = [ L.UUV1 ],
              couplings = {(0,0):C.GC_53})

V_20 = Vertex(name = 'V_20',
              particles = [ P.ghWm, P.ghZ__tilde__, P.G__plus__ ],
              color = [ '1' ],
              lorentz = [ L.UUS1 ],
              couplings = {(0,0):C.GC_78})

V_21 = Vertex(name = 'V_21',
              particles = [ P.ghWm, P.ghZ__tilde__, P.W__plus__ ],
              color = [ '1' ],
              lorentz = [ L.UUV1 ],
              couplings = {(0,0):C.GC_52})

V_22 = Vertex(name = 'V_22',
              particles = [ P.ghWp, P.ghA__tilde__, P.G__minus__ ],
              color = [ '1' ],
              lorentz = [ L.UUS1 ],
              couplings = {(0,0):C.GC_74})

V_23 = Vertex(name = 'V_23',
              particles = [ P.ghWp, P.ghA__tilde__, P.W__minus__ ],
              color = [ '1' ],
              lorentz = [ L.UUV1 ],
              couplings = {(0,0):C.GC_4})

V_24 = Vertex(name = 'V_24',
              particles = [ P.ghWp, P.ghWp__tilde__, P.G0 ],
              color = [ '1' ],
              lorentz = [ L.UUS1 ],
              couplings = {(0,0):C.GC_73})

V_25 = Vertex(name = 'V_25',
              particles = [ P.ghWp, P.ghWp__tilde__, P.H ],
              color = [ '1' ],
              lorentz = [ L.UUS1 ],
              couplings = {(0,0):C.GC_71})

V_26 = Vertex(name = 'V_26',
              particles = [ P.ghWp, P.ghWp__tilde__, P.a ],
              color = [ '1' ],
              lorentz = [ L.UUV1 ],
              couplings = {(0,0):C.GC_3})

V_27 = Vertex(name = 'V_27',
              particles = [ P.ghWp, P.ghWp__tilde__, P.Z ],
              color = [ '1' ],
              lorentz = [ L.UUV1 ],
              couplings = {(0,0):C.GC_52})

V_28 = Vertex(name = 'V_28',
              particles = [ P.ghWp, P.ghZ__tilde__, P.G__minus__ ],
              color = [ '1' ],
              lorentz = [ L.UUS1 ],
              couplings = {(0,0):C.GC_77})

V_29 = Vertex(name = 'V_29',
              particles = [ P.ghWp, P.ghZ__tilde__, P.W__minus__ ],
              color = [ '1' ],
              lorentz = [ L.UUV1 ],
              couplings = {(0,0):C.GC_53})

V_30 = Vertex(name = 'V_30',
              particles = [ P.ghZ, P.ghWm__tilde__, P.G__minus__ ],
              color = [ '1' ],
              lorentz = [ L.UUS1 ],
              couplings = {(0,0):C.GC_79})

V_31 = Vertex(name = 'V_31',
              particles = [ P.ghZ, P.ghWm__tilde__, P.W__minus__ ],
              color = [ '1' ],
              lorentz = [ L.UUV1 ],
              couplings = {(0,0):C.GC_52})

V_32 = Vertex(name = 'V_32',
              particles = [ P.ghZ, P.ghWp__tilde__, P.G__plus__ ],
              color = [ '1' ],
              lorentz = [ L.UUS1 ],
              couplings = {(0,0):C.GC_76})

V_33 = Vertex(name = 'V_33',
              particles = [ P.ghZ, P.ghWp__tilde__, P.W__plus__ ],
              color = [ '1' ],
              lorentz = [ L.UUV1 ],
              couplings = {(0,0):C.GC_53})

V_34 = Vertex(name = 'V_34',
              particles = [ P.ghZ, P.ghZ__tilde__, P.H ],
              color = [ '1' ],
              lorentz = [ L.UUS1 ],
              couplings = {(0,0):C.GC_80})

V_35 = Vertex(name = 'V_35',
              particles = [ P.ghG, P.ghG__tilde__, P.g ],
              color = [ 'f(3,1,2)' ],
              lorentz = [ L.UUV1 ],
              couplings = {(0,0):C.GC_10})

V_36 = Vertex(name = 'V_36',
              particles = [ P.g, P.g, P.g ],
              color = [ 'f(1,2,3)' ],
              lorentz = [ L.VVV1 ],
              couplings = {(0,0):C.GC_10})

V_37 = Vertex(name = 'V_37',
              particles = [ P.g, P.g, P.g, P.g ],
              color = [ 'f(-1,1,2)*f(3,4,-1)', 'f(-1,1,3)*f(2,4,-1)', 'f(-1,1,4)*f(2,3,-1)' ],
              lorentz = [ L.VVVV1, L.VVVV3, L.VVVV4 ],
              couplings = {(1,1):C.GC_12,(0,0):C.GC_12,(2,2):C.GC_12})

V_38 = Vertex(name = 'V_38',
              particles = [ P.a, P.W__minus__, P.G0, P.G__plus__ ],
              color = [ '1' ],
              lorentz = [ L.VVSS1 ],
              couplings = {(0,0):C.GC_55})

V_39 = Vertex(name = 'V_39',
              particles = [ P.a, P.W__minus__, P.G__plus__, P.H ],
              color = [ '1' ],
              lorentz = [ L.VVSS1 ],
              couplings = {(0,0):C.GC_54})

V_40 = Vertex(name = 'V_40',
              particles = [ P.a, P.W__minus__, P.G__plus__ ],
              color = [ '1' ],
              lorentz = [ L.VVS1 ],
              couplings = {(0,0):C.GC_74})

V_41 = Vertex(name = 'V_41',
              particles = [ P.W__minus__, P.G0, P.G__plus__ ],
              color = [ '1' ],
              lorentz = [ L.VSS1 ],
              couplings = {(0,0):C.GC_39})

V_42 = Vertex(name = 'V_42',
              particles = [ P.W__minus__, P.G__plus__, P.H ],
              color = [ '1' ],
              lorentz = [ L.VSS1 ],
              couplings = {(0,0):C.GC_37})

V_43 = Vertex(name = 'V_43',
              particles = [ P.a, P.W__minus__, P.W__plus__ ],
              color = [ '1' ],
              lorentz = [ L.VVV1 ],
              couplings = {(0,0):C.GC_4})

V_44 = Vertex(name = 'V_44',
              particles = [ P.a, P.W__plus__, P.G0, P.G__minus__ ],
              color = [ '1' ],
              lorentz = [ L.VVSS1 ],
              couplings = {(0,0):C.GC_55})

V_45 = Vertex(name = 'V_45',
              particles = [ P.a, P.W__plus__, P.G__minus__, P.H ],
              color = [ '1' ],
              lorentz = [ L.VVSS1 ],
              couplings = {(0,0):C.GC_56})

V_46 = Vertex(name = 'V_46',
              particles = [ P.a, P.W__plus__, P.G__minus__ ],
              color = [ '1' ],
              lorentz = [ L.VVS1 ],
              couplings = {(0,0):C.GC_75})

V_47 = Vertex(name = 'V_47',
              particles = [ P.W__plus__, P.G0, P.G__minus__ ],
              color = [ '1' ],
              lorentz = [ L.VSS1 ],
              couplings = {(0,0):C.GC_38})

V_48 = Vertex(name = 'V_48',
              particles = [ P.W__plus__, P.G__minus__, P.H ],
              color = [ '1' ],
              lorentz = [ L.VSS1 ],
              couplings = {(0,0):C.GC_37})

V_49 = Vertex(name = 'V_49',
              particles = [ P.W__minus__, P.W__plus__, P.G0, P.G0 ],
              color = [ '1' ],
              lorentz = [ L.VVSS1 ],
              couplings = {(0,0):C.GC_34})

V_50 = Vertex(name = 'V_50',
              particles = [ P.W__minus__, P.W__plus__, P.G__minus__, P.G__plus__ ],
              color = [ '1' ],
              lorentz = [ L.VVSS1 ],
              couplings = {(0,0):C.GC_34})

V_51 = Vertex(name = 'V_51',
              particles = [ P.W__minus__, P.W__plus__, P.H, P.H ],
              color = [ '1' ],
              lorentz = [ L.VVSS1 ],
              couplings = {(0,0):C.GC_34})

V_52 = Vertex(name = 'V_52',
              particles = [ P.W__minus__, P.W__plus__, P.H ],
              color = [ '1' ],
              lorentz = [ L.VVS1 ],
              couplings = {(0,0):C.GC_72})

V_53 = Vertex(name = 'V_53',
              particles = [ P.a, P.a, P.W__minus__, P.W__plus__ ],
              color = [ '1' ],
              lorentz = [ L.VVVV2 ],
              couplings = {(0,0):C.GC_5})

V_54 = Vertex(name = 'V_54',
              particles = [ P.W__minus__, P.W__plus__, P.Z ],
              color = [ '1' ],
              lorentz = [ L.VVV1 ],
              couplings = {(0,0):C.GC_53})

V_55 = Vertex(name = 'V_55',
              particles = [ P.W__minus__, P.W__minus__, P.W__plus__, P.W__plus__ ],
              color = [ '1' ],
              lorentz = [ L.VVVV2 ],
              couplings = {(0,0):C.GC_35})

V_56 = Vertex(name = 'V_56',
              particles = [ P.a, P.Z, P.G__minus__, P.G__plus__ ],
              color = [ '1' ],
              lorentz = [ L.VVSS1 ],
              couplings = {(0,0):C.GC_63})

V_57 = Vertex(name = 'V_57',
              particles = [ P.Z, P.G0, P.H ],
              color = [ '1' ],
              lorentz = [ L.VSS1 ],
              couplings = {(0,0):C.GC_60})

V_58 = Vertex(name = 'V_58',
              particles = [ P.Z, P.G__minus__, P.G__plus__ ],
              color = [ '1' ],
              lorentz = [ L.VSS1 ],
              couplings = {(0,0):C.GC_61})

V_59 = Vertex(name = 'V_59',
              particles = [ P.W__minus__, P.Z, P.G0, P.G__plus__ ],
              color = [ '1' ],
              lorentz = [ L.VVSS1 ],
              couplings = {(0,0):C.GC_8})

V_60 = Vertex(name = 'V_60',
              particles = [ P.W__minus__, P.Z, P.G__plus__, P.H ],
              color = [ '1' ],
              lorentz = [ L.VVSS1 ],
              couplings = {(0,0):C.GC_9})

V_61 = Vertex(name = 'V_61',
              particles = [ P.W__minus__, P.Z, P.G__plus__ ],
              color = [ '1' ],
              lorentz = [ L.VVS1 ],
              couplings = {(0,0):C.GC_67})

V_62 = Vertex(name = 'V_62',
              particles = [ P.W__plus__, P.Z, P.G0, P.G__minus__ ],
              color = [ '1' ],
              lorentz = [ L.VVSS1 ],
              couplings = {(0,0):C.GC_8})

V_63 = Vertex(name = 'V_63',
              particles = [ P.W__plus__, P.Z, P.G__minus__, P.H ],
              color = [ '1' ],
              lorentz = [ L.VVSS1 ],
              couplings = {(0,0):C.GC_7})

V_64 = Vertex(name = 'V_64',
              particles = [ P.W__plus__, P.Z, P.G__minus__ ],
              color = [ '1' ],
              lorentz = [ L.VVS1 ],
              couplings = {(0,0):C.GC_66})

V_65 = Vertex(name = 'V_65',
              particles = [ P.a, P.W__minus__, P.W__plus__, P.Z ],
              color = [ '1' ],
              lorentz = [ L.VVVV5 ],
              couplings = {(0,0):C.GC_57})

V_66 = Vertex(name = 'V_66',
              particles = [ P.Z, P.Z, P.G0, P.G0 ],
              color = [ '1' ],
              lorentz = [ L.VVSS1 ],
              couplings = {(0,0):C.GC_65})

V_67 = Vertex(name = 'V_67',
              particles = [ P.Z, P.Z, P.G__minus__, P.G__plus__ ],
              color = [ '1' ],
              lorentz = [ L.VVSS1 ],
              couplings = {(0,0):C.GC_64})

V_68 = Vertex(name = 'V_68',
              particles = [ P.Z, P.Z, P.H, P.H ],
              color = [ '1' ],
              lorentz = [ L.VVSS1 ],
              couplings = {(0,0):C.GC_65})

V_69 = Vertex(name = 'V_69',
              particles = [ P.Z, P.Z, P.H ],
              color = [ '1' ],
              lorentz = [ L.VVS1 ],
              couplings = {(0,0):C.GC_81})

V_70 = Vertex(name = 'V_70',
              particles = [ P.W__minus__, P.W__plus__, P.Z, P.Z ],
              color = [ '1' ],
              lorentz = [ L.VVVV2 ],
              couplings = {(0,0):C.GC_36})

V_71 = Vertex(name = 'V_71',
              particles = [ P.d__tilde__, P.d, P.a ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFV1 ],
              couplings = {(0,0):C.GC_1})

V_72 = Vertex(name = 'V_72',
              particles = [ P.s__tilde__, P.s, P.a ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFV1 ],
              couplings = {(0,0):C.GC_1})

V_73 = Vertex(name = 'V_73',
              particles = [ P.b__tilde__, P.b, P.a ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFV1 ],
              couplings = {(0,0):C.GC_1})

V_74 = Vertex(name = 'V_74',
              particles = [ P.d__tilde__, P.d, P.g ],
              color = [ 'T(3,2,1)' ],
              lorentz = [ L.FFV1 ],
              couplings = {(0,0):C.GC_11})

V_75 = Vertex(name = 'V_75',
              particles = [ P.s__tilde__, P.s, P.g ],
              color = [ 'T(3,2,1)' ],
              lorentz = [ L.FFV1 ],
              couplings = {(0,0):C.GC_11})

V_76 = Vertex(name = 'V_76',
              particles = [ P.b__tilde__, P.b, P.g ],
              color = [ 'T(3,2,1)' ],
              lorentz = [ L.FFV1 ],
              couplings = {(0,0):C.GC_11})

V_77 = Vertex(name = 'V_77',
              particles = [ P.b__tilde__, P.b, P.G0 ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFS2 ],
              couplings = {(0,0):C.GC_82})

V_78 = Vertex(name = 'V_78',
              particles = [ P.b__tilde__, P.b, P.H ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFS4 ],
              couplings = {(0,0):C.GC_83})

V_79 = Vertex(name = 'V_79',
              particles = [ P.d__tilde__, P.d, P.Z ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFV2, L.FFV3 ],
              couplings = {(0,0):C.GC_50,(0,1):C.GC_58})

V_80 = Vertex(name = 'V_80',
              particles = [ P.s__tilde__, P.s, P.Z ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFV2, L.FFV3 ],
              couplings = {(0,0):C.GC_50,(0,1):C.GC_58})

V_81 = Vertex(name = 'V_81',
              particles = [ P.b__tilde__, P.b, P.Z ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFV2, L.FFV3 ],
              couplings = {(0,1):C.GC_58,(0,0):C.GC_50})

V_82 = Vertex(name = 'V_82',
              particles = [ P.c__tilde__, P.d, P.G__plus__ ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFS3 ],
              couplings = {(0,0):C.GC_16})

V_83 = Vertex(name = 'V_83',
              particles = [ P.t__tilde__, P.d, P.G__plus__ ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFS3 ],
              couplings = {(0,0):C.GC_17})

V_84 = Vertex(name = 'V_84',
              particles = [ P.c__tilde__, P.s, P.G__plus__ ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFS3 ],
              couplings = {(0,0):C.GC_18})

V_85 = Vertex(name = 'V_85',
              particles = [ P.t__tilde__, P.s, P.G__plus__ ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFS3 ],
              couplings = {(0,0):C.GC_19})

V_86 = Vertex(name = 'V_86',
              particles = [ P.u__tilde__, P.b, P.G__plus__ ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFS1 ],
              couplings = {(0,0):C.GC_13})

V_87 = Vertex(name = 'V_87',
              particles = [ P.c__tilde__, P.b, P.G__plus__ ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFS1, L.FFS3 ],
              couplings = {(0,0):C.GC_14,(0,1):C.GC_20})

V_88 = Vertex(name = 'V_88',
              particles = [ P.t__tilde__, P.b, P.G__plus__ ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFS1, L.FFS3 ],
              couplings = {(0,0):C.GC_15,(0,1):C.GC_21})

V_89 = Vertex(name = 'V_89',
              particles = [ P.u__tilde__, P.d, P.W__plus__ ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFV2 ],
              couplings = {(0,0):C.GC_100})

V_90 = Vertex(name = 'V_90',
              particles = [ P.c__tilde__, P.d, P.W__plus__ ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFV2 ],
              couplings = {(0,0):C.GC_103})

V_91 = Vertex(name = 'V_91',
              particles = [ P.t__tilde__, P.d, P.W__plus__ ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFV2 ],
              couplings = {(0,0):C.GC_106})

V_92 = Vertex(name = 'V_92',
              particles = [ P.u__tilde__, P.s, P.W__plus__ ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFV2 ],
              couplings = {(0,0):C.GC_101})

V_93 = Vertex(name = 'V_93',
              particles = [ P.c__tilde__, P.s, P.W__plus__ ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFV2 ],
              couplings = {(0,0):C.GC_104})

V_94 = Vertex(name = 'V_94',
              particles = [ P.t__tilde__, P.s, P.W__plus__ ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFV2 ],
              couplings = {(0,0):C.GC_107})

V_95 = Vertex(name = 'V_95',
              particles = [ P.u__tilde__, P.b, P.W__plus__ ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFV2 ],
              couplings = {(0,0):C.GC_102})

V_96 = Vertex(name = 'V_96',
              particles = [ P.c__tilde__, P.b, P.W__plus__ ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFV2 ],
              couplings = {(0,0):C.GC_105})

V_97 = Vertex(name = 'V_97',
              particles = [ P.t__tilde__, P.b, P.W__plus__ ],
              color = [ 'Identity(1,2)' ],
              lorentz = [ L.FFV2 ],
              couplings = {(0,0):C.GC_108})

V_98 = Vertex(name = 'V_98',
              particles = [ P.e__plus__, P.e__minus__, P.a ],
              color = [ '1' ],
              lorentz = [ L.FFV1 ],
              couplings = {(0,0):C.GC_3})

V_99 = Vertex(name = 'V_99',
              particles = [ P.mu__plus__, P.mu__minus__, P.a ],
              color = [ '1' ],
              lorentz = [ L.FFV1 ],
              couplings = {(0,0):C.GC_3})

V_100 = Vertex(name = 'V_100',
               particles = [ P.ta__plus__, P.ta__minus__, P.a ],
               color = [ '1' ],
               lorentz = [ L.FFV1 ],
               couplings = {(0,0):C.GC_3})

V_101 = Vertex(name = 'V_101',
               particles = [ P.e__plus__, P.e__minus__, P.G0 ],
               color = [ '1' ],
               lorentz = [ L.FFS2 ],
               couplings = {(0,0):C.GC_88})

V_102 = Vertex(name = 'V_102',
               particles = [ P.mu__plus__, P.mu__minus__, P.G0 ],
               color = [ '1' ],
               lorentz = [ L.FFS2 ],
               couplings = {(0,0):C.GC_92})

V_103 = Vertex(name = 'V_103',
               particles = [ P.ta__plus__, P.ta__minus__, P.G0 ],
               color = [ '1' ],
               lorentz = [ L.FFS2 ],
               couplings = {(0,0):C.GC_98})

V_104 = Vertex(name = 'V_104',
               particles = [ P.e__plus__, P.e__minus__, P.H ],
               color = [ '1' ],
               lorentz = [ L.FFS4 ],
               couplings = {(0,0):C.GC_89})

V_105 = Vertex(name = 'V_105',
               particles = [ P.mu__plus__, P.mu__minus__, P.H ],
               color = [ '1' ],
               lorentz = [ L.FFS4 ],
               couplings = {(0,0):C.GC_93})

V_106 = Vertex(name = 'V_106',
               particles = [ P.ta__plus__, P.ta__minus__, P.H ],
               color = [ '1' ],
               lorentz = [ L.FFS4 ],
               couplings = {(0,0):C.GC_99})

V_107 = Vertex(name = 'V_107',
               particles = [ P.e__plus__, P.e__minus__, P.Z ],
               color = [ '1' ],
               lorentz = [ L.FFV2, L.FFV4 ],
               couplings = {(0,0):C.GC_50,(0,1):C.GC_59})

V_108 = Vertex(name = 'V_108',
               particles = [ P.mu__plus__, P.mu__minus__, P.Z ],
               color = [ '1' ],
               lorentz = [ L.FFV2, L.FFV4 ],
               couplings = {(0,0):C.GC_50,(0,1):C.GC_59})

V_109 = Vertex(name = 'V_109',
               particles = [ P.ta__plus__, P.ta__minus__, P.Z ],
               color = [ '1' ],
               lorentz = [ L.FFV2, L.FFV4 ],
               couplings = {(0,0):C.GC_50,(0,1):C.GC_59})

V_110 = Vertex(name = 'V_110',
               particles = [ P.ve__tilde__, P.e__minus__, P.G__plus__ ],
               color = [ '1' ],
               lorentz = [ L.FFS1 ],
               couplings = {(0,0):C.GC_87})

V_111 = Vertex(name = 'V_111',
               particles = [ P.vm__tilde__, P.mu__minus__, P.G__plus__ ],
               color = [ '1' ],
               lorentz = [ L.FFS1 ],
               couplings = {(0,0):C.GC_91})

V_112 = Vertex(name = 'V_112',
               particles = [ P.vt__tilde__, P.ta__minus__, P.G__plus__ ],
               color = [ '1' ],
               lorentz = [ L.FFS1 ],
               couplings = {(0,0):C.GC_97})

V_113 = Vertex(name = 'V_113',
               particles = [ P.ve__tilde__, P.e__minus__, P.W__plus__ ],
               color = [ '1' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_40})

V_114 = Vertex(name = 'V_114',
               particles = [ P.vm__tilde__, P.mu__minus__, P.W__plus__ ],
               color = [ '1' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_40})

V_115 = Vertex(name = 'V_115',
               particles = [ P.vt__tilde__, P.ta__minus__, P.W__plus__ ],
               color = [ '1' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_40})

V_116 = Vertex(name = 'V_116',
               particles = [ P.b__tilde__, P.u, P.G__minus__ ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFS3 ],
               couplings = {(0,0):C.GC_28})

V_117 = Vertex(name = 'V_117',
               particles = [ P.d__tilde__, P.c, P.G__minus__ ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFS1 ],
               couplings = {(0,0):C.GC_22})

V_118 = Vertex(name = 'V_118',
               particles = [ P.s__tilde__, P.c, P.G__minus__ ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFS1 ],
               couplings = {(0,0):C.GC_23})

V_119 = Vertex(name = 'V_119',
               particles = [ P.b__tilde__, P.c, P.G__minus__ ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFS1, L.FFS3 ],
               couplings = {(0,0):C.GC_24,(0,1):C.GC_29})

V_120 = Vertex(name = 'V_120',
               particles = [ P.d__tilde__, P.t, P.G__minus__ ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFS1 ],
               couplings = {(0,0):C.GC_25})

V_121 = Vertex(name = 'V_121',
               particles = [ P.s__tilde__, P.t, P.G__minus__ ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFS1 ],
               couplings = {(0,0):C.GC_26})

V_122 = Vertex(name = 'V_122',
               particles = [ P.b__tilde__, P.t, P.G__minus__ ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFS1, L.FFS3 ],
               couplings = {(0,0):C.GC_27,(0,1):C.GC_30})

V_123 = Vertex(name = 'V_123',
               particles = [ P.d__tilde__, P.u, P.W__minus__ ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_41})

V_124 = Vertex(name = 'V_124',
               particles = [ P.s__tilde__, P.u, P.W__minus__ ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_42})

V_125 = Vertex(name = 'V_125',
               particles = [ P.b__tilde__, P.u, P.W__minus__ ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_43})

V_126 = Vertex(name = 'V_126',
               particles = [ P.d__tilde__, P.c, P.W__minus__ ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_44})

V_127 = Vertex(name = 'V_127',
               particles = [ P.s__tilde__, P.c, P.W__minus__ ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_45})

V_128 = Vertex(name = 'V_128',
               particles = [ P.b__tilde__, P.c, P.W__minus__ ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_46})

V_129 = Vertex(name = 'V_129',
               particles = [ P.d__tilde__, P.t, P.W__minus__ ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_47})

V_130 = Vertex(name = 'V_130',
               particles = [ P.s__tilde__, P.t, P.W__minus__ ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_48})

V_131 = Vertex(name = 'V_131',
               particles = [ P.b__tilde__, P.t, P.W__minus__ ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_49})

V_132 = Vertex(name = 'V_132',
               particles = [ P.u__tilde__, P.u, P.a ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFV1 ],
               couplings = {(0,0):C.GC_2})

V_133 = Vertex(name = 'V_133',
               particles = [ P.c__tilde__, P.c, P.a ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFV1 ],
               couplings = {(0,0):C.GC_2})

V_134 = Vertex(name = 'V_134',
               particles = [ P.t__tilde__, P.t, P.a ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFV1 ],
               couplings = {(0,0):C.GC_2})

V_135 = Vertex(name = 'V_135',
               particles = [ P.u__tilde__, P.u, P.g ],
               color = [ 'T(3,2,1)' ],
               lorentz = [ L.FFV1 ],
               couplings = {(0,0):C.GC_11})

V_136 = Vertex(name = 'V_136',
               particles = [ P.c__tilde__, P.c, P.g ],
               color = [ 'T(3,2,1)' ],
               lorentz = [ L.FFV1 ],
               couplings = {(0,0):C.GC_11})

V_137 = Vertex(name = 'V_137',
               particles = [ P.t__tilde__, P.t, P.g ],
               color = [ 'T(3,2,1)' ],
               lorentz = [ L.FFV1 ],
               couplings = {(0,0):C.GC_11})

V_138 = Vertex(name = 'V_138',
               particles = [ P.c__tilde__, P.c, P.G0 ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFS2 ],
               couplings = {(0,0):C.GC_85})

V_139 = Vertex(name = 'V_139',
               particles = [ P.t__tilde__, P.t, P.G0 ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFS2 ],
               couplings = {(0,0):C.GC_95})

V_140 = Vertex(name = 'V_140',
               particles = [ P.c__tilde__, P.c, P.H ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFS4 ],
               couplings = {(0,0):C.GC_84})

V_141 = Vertex(name = 'V_141',
               particles = [ P.t__tilde__, P.t, P.H ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFS4 ],
               couplings = {(0,0):C.GC_94})

V_142 = Vertex(name = 'V_142',
               particles = [ P.u__tilde__, P.u, P.Z ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFV2, L.FFV5 ],
               couplings = {(0,0):C.GC_51,(0,1):C.GC_58})

V_143 = Vertex(name = 'V_143',
               particles = [ P.c__tilde__, P.c, P.Z ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFV2, L.FFV5 ],
               couplings = {(0,0):C.GC_51,(0,1):C.GC_58})

V_144 = Vertex(name = 'V_144',
               particles = [ P.t__tilde__, P.t, P.Z ],
               color = [ 'Identity(1,2)' ],
               lorentz = [ L.FFV2, L.FFV5 ],
               couplings = {(0,0):C.GC_51,(0,1):C.GC_58})

V_145 = Vertex(name = 'V_145',
               particles = [ P.e__plus__, P.ve, P.G__minus__ ],
               color = [ '1' ],
               lorentz = [ L.FFS3 ],
               couplings = {(0,0):C.GC_86})

V_146 = Vertex(name = 'V_146',
               particles = [ P.mu__plus__, P.vm, P.G__minus__ ],
               color = [ '1' ],
               lorentz = [ L.FFS3 ],
               couplings = {(0,0):C.GC_90})

V_147 = Vertex(name = 'V_147',
               particles = [ P.ta__plus__, P.vt, P.G__minus__ ],
               color = [ '1' ],
               lorentz = [ L.FFS3 ],
               couplings = {(0,0):C.GC_96})

V_148 = Vertex(name = 'V_148',
               particles = [ P.e__plus__, P.ve, P.W__minus__ ],
               color = [ '1' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_40})

V_149 = Vertex(name = 'V_149',
               particles = [ P.mu__plus__, P.vm, P.W__minus__ ],
               color = [ '1' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_40})

V_150 = Vertex(name = 'V_150',
               particles = [ P.ta__plus__, P.vt, P.W__minus__ ],
               color = [ '1' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_40})

V_151 = Vertex(name = 'V_151',
               particles = [ P.ve__tilde__, P.ve, P.Z ],
               color = [ '1' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_62})

V_152 = Vertex(name = 'V_152',
               particles = [ P.vm__tilde__, P.vm, P.Z ],
               color = [ '1' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_62})

V_153 = Vertex(name = 'V_153',
               particles = [ P.vt__tilde__, P.vt, P.Z ],
               color = [ '1' ],
               lorentz = [ L.FFV2 ],
               couplings = {(0,0):C.GC_62})




# type: ignore
from __future__ import absolute_import
from .object_library import all_orders, CouplingOrder

# Keep the coupling order names QCD and QED as they are hard-coded standards in some generators
QCD = CouplingOrder(name='QCD',
                    expansion_order=99,
                    hierarchy=1)

QED = CouplingOrder(name='QED',
                    expansion_order=99,
                    hierarchy=2)
